"""
This script is run on the Python command line or on the equivalent `main`
function to execute data enrichment. If a custom enricher is used, then the
`enrich_dataset` function is used.
"""
################################################################################


import argparse
import datetime
import os
import pathlib
import shutil
import sys
from typing import List, Type
from uuid import uuid4
from watchful import client, attributes
from watchful.enricher import Enricher


def enrich_dataset(
    custom_enricher_cls: Type[Enricher], args: List[str] = None
) -> None:
    """
    This is the function to use for performing custom data enrichment. Custom
    data enrichment variables and functions defined in custom_enricher_cls
    are used to perform the data enrichment.
    """
    if args is None:
        args = []
    attributes.set_multiprocessing(False)
    custom_enricher = custom_enricher_cls()

    main(args, custom_enricher)


def main(args: List[str] = None, custom_enricher: Enricher = None) -> None:
    if args is None:
        args = []

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # The dataset file; if not provided use the dataset currently opened in
    # Watchful. Not used in remote enrichment as we will retrieve the dataset
    # from the remote Watchful application.
    parser.add_argument(
        "--in_file",
        type=str,
        default="",
        help="Optional original csv dataset filepath; if not given use the \
            current dataset opened in watchfu.",
    )
    # This is the attributes output file that is specifically formatted for
    # integration with the Watchful application.
    parser.add_argument(
        "--out_file",
        type=str,
        default="",
        help='Optional output attribute filepath; if given must end with \
            ".attrs" extension.',
    )
    # The attribute file to ingest from, if it is available.
    parser.add_argument(
        "--attr_file",
        type=str,
        default="",
        help="Optional input csv attribute filepath; if not given create the \
            initial spacy attributes.",
    )
    # Columns in the `attr_file` csv file to use as attributes; use all
    # attributes if not provided.
    parser.add_argument(
        "--attr_names",
        type=str,
        default="",
        help="Optional comma delimited string of attribute names to be used; \
            if not given use all attribute names.",
    )
    # The host to use; if not provided use localhost.
    parser.add_argument(
        "--wf_host",
        type=str,
        default="localhost",
        help='Optional string host running Watchful; if not given use \
            "localhost".',
    )
    # The port to use; if not provided use port 9001.
    parser.add_argument(
        "--wf_port",
        type=str,
        default="9001",
        help='Optional string port number running Watchful; if not given use \
            "9001".',
    )
    # The out-of-the-box NLP to use if no `attr_file` is provided.
    parser.add_argument(
        "--standard_nlp",
        type=str,
        default="spacy",
        help='Optional out-of-the-box NLP to use, currently "spacy" and \
            "flair" are available; "spacy" if unspecified.',
    )
    parser.add_argument(
        "--multiprocessing",
        action="store_true",
        help="Optional explicit use of multiprocessing on available physical \
            cpu cores; no explicit use if unspecified.",
    )
    parser.add_argument(
        "--is_local",
        action="store_true",
        help="Optional use of local watchful client application instead of a \
            hosted application; hosted application if unspecified.",
    )
    args = parser.parse_args(args=args)

    attributes.set_multiprocessing(args.multiprocessing)

    client.external(host=args.wf_host, port=args.wf_port)

    summary = client.get()
    project_id = client.get_project_id(summary)
    (
        dataset_id,
        datasets_dir,
        args.in_file,
    ) = attributes.get_dataset_id_dir_filepath(
        summary, args.in_file, args.is_local
    )

    # `args.in_file` will still be returned as "" if Watchful application is
    # remote. Therefore, we create a temporary filepath for `args.in_file` to
    # download the original dataset to.
    if not args.is_local:
        if args.in_file == "":
            user_home_path = os.path.expanduser("~")
            working_dir = os.path.join(user_home_path, "watchful", "working")
            os.makedirs(working_dir, exist_ok=True)
            args.in_file = os.path.join(working_dir, str(uuid4()))
            client.export_dataset_to(
                args.in_file, fields=summary["field_names"]
            )
        else:
            print(
                'in_file must be initially "" for enrichment to a remote '
                f'Watchful applcation; got "{args.out_file}" instead.'
            )
            sys.exit(1)

    if args.out_file:
        # Check that `out_file` has ".attrs" extension.
        try:
            is_ext_attrs = os.path.splitext(args.out_file)[1] == ".attrs"
            if is_ext_attrs:
                del_out_file = False
                # check that out_file's directory exists
                out_file_dir = pathlib.Path(args.out_file).parent
                if not os.path.isdir(out_file_dir):
                    print(f"Directory {out_file_dir} does not exist.")
                    sys.exit(1)
            else:
                print(
                    f'out_file {args.out_file} must end with ".attrs" '
                    "extension."
                )
                sys.exit(1)
        except OSError as err_msg:
            print(err_msg)
            print(f'out_file {args.out_file} must end with ".attrs" extension.')
            sys.exit(1)
    else:
        del_out_file = True
        # Create a temporary `out_file` and mark it for deletion.
        args.out_file = f"{os.path.splitext(args.in_file)[0]}.attrs"

    # Enrich with attributes from a csv file, that is, already created from an
    # external pipeline.
    if args.attr_file:

        # Enrich with all attributes.
        if not args.attr_names:
            print(
                f"Enriching {args.in_file} using all attributes from "
                f"{args.attr_file} ..."
            )
            attributes.enrich(
                args.in_file,
                args.out_file,
                attributes.enrich_row_with_attribute_data,
                attributes.get_vars_for_enrich_row_with_attribute_data(
                    args.attr_names, args.attr_file
                ),
            )
            if not del_out_file:
                print(f"Wrote attributes to {args.out_file}.")

        # Enrich with specified attributes.
        else:
            val_success = attributes.validate_attr_names(
                args.attr_names, args.attr_file
            )

            if val_success:
                print(
                    f"Enriching {args.in_file} using {args.attr_names} "
                    f"attributes from {args.attr_file} ..."
                )
                attributes.enrich(
                    args.in_file,
                    args.out_file,
                    attributes.enrich_row_with_attribute_data,
                    attributes.get_vars_for_enrich_row_with_attribute_data(
                        args.attr_names, args.attr_file
                    ),
                )
                if not del_out_file:
                    print(f"Wrote attributes to {args.out_file}.")
            else:
                print(
                    f"At least one of your attribute names in "
                    f"{args.attr_names} do not match those in the attribute "
                    f"input file {args.attr_file}."
                )
                sys.exit(1)

    # Do data enrichment from standard NLPs.
    elif custom_enricher is None:

        # Do SpaCy.
        if args.standard_nlp == "spacy":
            # Want to know what pipes are used? Uncomment these:
            # nlp = attributes.load_spacy()
            # import pprint
            # pprint.PrettyPrinter(indent=4).pprint(nlp.analyze_pipes())

            # `enrich_row` is the user custom function for enriching every row
            # of the dataset. `spacy_atterize_fn`, `spacy_atterize` and
            # `load_spacy()` are the additional user variables to perform the
            # data enrichment.
            print(f"Using {args.standard_nlp} ...")
            print(f"Enriching {args.in_file} ...")
            attributes.enrich(
                args.in_file,
                args.out_file,
                attributes.enrich_row,
                (
                    attributes.spacy_atterize_fn,
                    attributes.spacy_atterize,
                    attributes.load_spacy(),
                ),
            )
            if not del_out_file:
                print(f"Wrote attributes to {args.out_file}.")

        # Do Flair.
        elif args.standard_nlp == "flair":
            # `enrich_row` is the user custom function for enriching every row
            # of the dataset. `flair_atterize_fn`, `flair_atterize` and
            # `*load_flair()` are the additional user variables to perform the
            # data enrichment.
            print(f"Using {args.standard_nlp} ...")
            print(f"Enriching {args.in_file} ...")
            attributes.enrich(
                args.in_file,
                args.out_file,
                attributes.enrich_row,
                (
                    attributes.flair_atterize_fn,
                    attributes.flair_atterize,
                    *attributes.load_flair(),
                ),
            )
            if not del_out_file:
                print(f"Wrote attributes to {args.out_file}.")

        else:
            print(
                f"The nlp {args.standard_nlp} is not implemented.\nNo "
                "enrichment done."
            )
            sys.exit(1)

    # Do custom data enrichment.
    else:
        # Perform custom data enrichment. Custom data enrichment variables and
        # functions in custom_enricher are used to perform the data enrichment.
        print("Using your custom enricher ...")
        print(f"Enriching {args.in_file} ...")
        attributes.enrich(
            args.in_file,
            args.out_file,
            custom_enricher.enrich_row,
            custom_enricher.enrichment_args,
        )
        if not del_out_file:
            print(f"Wrote attributes to {args.out_file}.")

    # If Watchful application is remote, delete the downloaded dataset as the
    # data enrichment is completed.
    if not args.is_local:
        try:
            os.remove(args.in_file)
        except FileNotFoundError as err_msg:
            print(err_msg)
            print(
                f"Error removing downloaded dataset file from {args.in_file}."
            )
            sys.exit(1)

    # Check that the active project and the opened dataset have not changed.
    summary = client.get()
    current_project_id = client.get_project_id(summary)
    if project_id != current_project_id:
        print(
            f"Current project {current_project_id} is different from the "
            f"enriched project {project_id}!"
        )
        sys.exit(1)

    current_dataset_id = client.get_dataset_id(summary)
    if dataset_id != current_dataset_id:
        print(
            f"Current dataset {current_dataset_id} is different from the "
            f"enriched dataset {dataset_id}!"
        )
        sys.exit(1)

    # Format the attributes file timestamp as "yyyy-mm-dd_hh-mm-ss-ssssss".
    # Use the full timestamp for completeness; though it's reasonable
    # "yyyy-mm-dd_hh-mm-ss" should work uniquely too.
    # This timestamp format is:
    #     - usable for pulling out the latest attributes
    #     - usable for pulling out a time-series progression of attributes
    #     - consistent here and in kubeflow pipelines integration
    # The final filename format is "filename__yyyy-mm-dd_hh-mm-ss-ssssss.attrs".
    timestamp = (
        str(datetime.datetime.now())
        .replace(" ", "_")
        .replace(":", "-")
        .replace(".", "-")
    )
    dest_attr_filename = f"__{timestamp}.".join(
        os.path.basename(args.out_file).split(".")
    )

    # Copy created attributes output file to watchful home attributes directory
    # if the watchful application is local.
    if args.is_local:
        dest_attr_filepath = os.path.join(
            datasets_dir, "attrs", dest_attr_filename
        )
        try:
            os.makedirs(os.path.dirname(dest_attr_filepath), exist_ok=True)
            shutil.copyfile(args.out_file, dest_attr_filepath)
        except OSError as err_msg:
            print(err_msg)
            print(
                f"Error copying attribute output file from {args.out_file} "
                f"to {dest_attr_filepath}."
            )
            sys.exit(1)
    # Upload created attributes output file to Watchful application if it is
    # remote.
    else:
        client.upload_attributes(dataset_id, args.out_file)

    if del_out_file:
        try:
            os.remove(args.out_file)
        except FileNotFoundError as err_msg:
            print(err_msg)
            print(
                "Error removing temporary attribute output file from "
                f"{args.out_file}."
            )
            sys.exit(1)

    if args.is_local:
        # Load attributes filepath into Watchful application.
        # Example usage:
        # curl -iX POST http://localhost:9001/api \
        #   --header "Content-Type: application/json" \
        #   --data '{"verb":"attributes","id":"9570b0b5-4a58-445f-9b51-b434caca2650",
        #         "filepath":"/path/to/attributes_file.attrs"}'
        # Arguments:
        #     id: dataset id
        #     filepath: attributes filepath
        load_attrib_res = client.load_attributes(dataset_id, dest_attr_filename)

        msg = (
            f"attributes via watchful {args.wf_host}:{args.wf_port} to dataset "
            f"id {dataset_id}."
        )
        if "error_msg" in load_attrib_res and load_attrib_res["error_msg"]:
            print(f"Error ingesting {msg}")
        else:
            print(f"Ingested {msg}")


if __name__ == "__main__":
    main(sys.argv[1:])



"""
This script provides the functions required for data enrichment.
"""
################################################################################


import csv
import json
import numbers
import os
import pprint
import re
import sys
from typing import Callable, Dict, List, Optional, Tuple, Union
from multiprocessing import Pool
import psutil
from watchful import client

pprint._sorted = lambda x: x
pprint = pprint.PrettyPrinter(indent=4).pprint


# Generally, you should not need to edit these variables directly.
IS_MULTIPROC = False
MULTIPROC_CHUNKSIZE = None
ENRICHMENT_ARGS = None
ATTR_WRITER = None
EnrichedCell = List[
    Tuple[Union[List[Tuple[int]], Dict[str, List[str]], Optional[str]]]
]

# Constants for encoding spans into compact strings
BASE = 64
COMP_LEN = 8

# '0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmno'
numerals = dict(
    map(lambda ic: (ic[0], chr(ic[1])), enumerate(range(48, 48 + BASE)))
)

# "#$%&'()*"
compressed = dict(
    map(lambda ic: (ic[0], chr(ic[1])), enumerate(range(35, 35 + COMP_LEN)))
)


def set_multiprocessing(is_multiproc: bool) -> None:
    global IS_MULTIPROC
    global MULTIPROC_CHUNKSIZE
    if is_multiproc and not IS_MULTIPROC:
        IS_MULTIPROC = True
        MULTIPROC_CHUNKSIZE = 500
    elif not is_multiproc:
        IS_MULTIPROC = False
        MULTIPROC_CHUNKSIZE = None


def set_multiproc_chunksize(multiproc_chunksize: int) -> None:
    assert multiproc_chunksize >= 1
    if IS_MULTIPROC:
        global MULTIPROC_CHUNKSIZE
        MULTIPROC_CHUNKSIZE = multiproc_chunksize


def print_multiproc_params() -> None:
    print(
        f"Multiprocessing: {IS_MULTIPROC}, "
        f"Multiprocessing chunksize: {MULTIPROC_CHUNKSIZE}"
    )


def base64(num):
    if num == 0:
        return numerals[0]

    ret = ""
    while num > 0:
        ret += numerals[num % BASE]
        num = num // BASE
    return ret[::-1]


def base64str(coll):
    """
    Takes a collection of numbers and returns a collection with strings
    representing those numbers in base64.

    Additional compression is done by concatenating consecutive base64 strings
    of the same length. This compressed encoding is detected by inspecting the
    first character in the ASCII range 35 inclusive to 42 inclusive. The rest of
    the string should be partitioned into their own strings with length
    ascii_code(s[0]) - 34.

    For example:
    * String: "#1234" represents "1,2,3,4"
    * String: "$1234" represents "12,34"
    * String: "&1234" represents "1234", (which is never compressed since the
    value is smaller than the compressed)
    (compression will not be done for a base64 encoded number if it is preceded
    or suceeded by a base64 encoded number of a different length)

    Here are the ascii codes for the prefixes in the examples above:
    * "#" => 35
    * "$" => 36
    * "&" => 38
    The range 35 inclusive to 42 inclusive was chosen because it contains
    characters that do not need to be escaped in JSON, nor does the range
    contain comma (','), which is used as a delimiter to concat all of the
    strings.
    """

    ret = []
    buf = []

    def push_buf(buf, s):
        if len(buf) == 0 or len(s) == len(buf[0]):
            buf.append(s)
        else:
            flush_buf()
            buf = push_buf([], s)
        return buf

    def flush_buf():
        buf_len = len(buf)
        if buf_len == 1:
            ret.append(buf[0])
        elif buf_len > 1:
            compress_idx = len(buf[0]) - 1
            if compress_idx < len(compressed):  # compressed limit
                ret.append(compressed[compress_idx] + "".join(buf))
            else:
                for x in buf:
                    ret.append(x)

    for x in coll:
        buf = push_buf(buf, base64(x))
    flush_buf()
    return ",".join(ret)


def contig_spans(spans):
    """
    Decode spans to contig_spans/2 with
    [gap_len,span_len,...,gap_len_N,span_len_N].
    Takes a list of (start, end).
    """

    contig = []
    offset = 0
    for (a, b) in spans:
        contig.append(a - offset)
        contig.append(b - a)
        offset = b
    return contig


def writer(out, n_rows, n_cols):
    """
    Takes the output file and the number of rows & columns of the original
    dataset. Returns a write function that takes all of the attributes for a
    cell in the original dataset, where a cell is a single row and column pair.

    The cells attributes should be in this shape:
    [(spans:Vec<(int,int)>, attr_vals:Map<String,Vec<Any>>, name:Option<String>)
    ..
    ]
    * `spans` is a sorted vector of span [start, end) in the cell.
    * `attr_vals` is a map from attribute name to values for the `spans`. None
    means that the attribute has no value for that token defined by its span.
    * `name` is an optional parameter which can be used to give a name to the
    spans, where the attribute value of that name is the content of the spans
    themselves. Examples of this is: sentences, noun_chunks, tokens or
    collage_names.
    """

    # Write attribute and value mappings
    attrs = {}
    values = {}

    def write_jsonl(obj):
        json.dump(obj, out, separators=(",", ":"))
        out.write("\n")

    def write(cell_data):
        new_attrs = []
        new_values = {}
        cell = []

        # For each span data tuple for the cell with the span and attribute
        # values.
        # for those span, do:
        for span_data in cell_data:
            span = span_data[0]
            attr_vals = span_data[1]
            name = span_data[2] if len(span_data) == 3 else None

            # Gather the new attributes and values
            # Gather and create new mappings at the same time. Duh :)
            for (attr, vals) in attr_vals.items():
                if attr not in attrs:
                    attrs[attr] = len(attrs) + 1
                    values[attr] = {}
                    new_attrs.append(attr)
                for val in vals:
                    if isinstance(val, numbers.Number):
                        val = str(val)
                    elif val and not isinstance(val, str):
                        raise Exception(
                            "Attribute value must be a string, "
                            "None or a number. Was: " + val
                        )
                    if val and not val in values[attr]:
                        values[attr][val] = len(values[attr]) + 1
                        if attr not in new_values:
                            new_values[attr] = []
                        new_values[attr].append(val)

            # Create the vector for the current cell
            span_val = []
            if name is not None:
                span_val.append(name)
            for (attr, vals) in attr_vals.items():
                assert len(span) == len(
                    vals
                ), "Must be the same amount of spans as attribute values."
                # Not base64 the attributes to save space since there aren't
                # that many of them.
                span_val.append(attrs[attr])
                span_val.append(
                    base64str([values[attr][val] if val else 0 for val in vals])
                )

            cell.append(base64str(contig_spans(span)))
            cell.append(span_val)

        # Output the lines (attributes, values and the cell itself)
        if new_attrs:
            write_jsonl(["@"] + new_attrs)
        if new_values:
            for (k, vals) in new_values.items():
                write_jsonl(["$", k] + vals)
        write_jsonl(cell)

    # Write the header and return the write function to be called by users
    write_jsonl({"version": "0.3", "rows": n_rows, "cols": n_cols})
    return write


def spacy_atterize(doc):
    """
    Extracts attributes from a SpaCy document. Attributes are associated to
    substrings, being tokens, entities, sentences or noun chunks. Every
    Substring is identified by its character start index and character end
    index.
    """

    # Return value
    cell = []

    def float_to_str(f):
        return str(round(f * 100.0))

    # Polarity and subjectivity via nlp.add_pipe("spacytextblob")
    cell.append(
        (
            [(0, len(doc.text))],
            {
                "polarity": [float_to_str(doc._.blob.polarity)],
                "subjectivity": [float_to_str(doc._.blob.subjectivity)],
            },
        )
    )

    # Token level assessment of the polarity and subjectivity
    assessments = {}
    for (
        words,
        polarity,
        subjectivity,
        _,
    ) in doc._.blob.sentiment_assessments.assessments:
        for word in words:
            regex = re.compile(f"\\b{re.escape(word)}\\b")
            for m in regex.finditer(doc.text):
                span = m.span()
                if not span in assessments:
                    assessments[span] = (1, [polarity], [subjectivity])
                else:
                    (c, ps, ss) = assessments.get(span)
                    ps.append(polarity)
                    ss.append(subjectivity)
                    assessments[span] = (c + 1, ps, ss)
    assess_spans = []
    pol_vals = []
    subj_vals = []
    for span in sorted(assessments.keys()):
        (c, ps, ss) = assessments.get(span)
        assess_spans.append(span)
        pol_vals.append(float_to_str(sum(ps) / c))
        subj_vals.append(float_to_str(sum(ss) / c))
    cell.append(
        (
            assess_spans,
            {"token_polarity": pol_vals, "token_subjectivity": subj_vals},
        )
    )

    # Token level attributes
    tok_spans = []
    pos_vals = []
    tag_vals = []
    lem_vals = []
    case_vals = []
    for token in doc:
        tok_spans.append((token.idx, token.idx + len(token.text)))
        pos_vals.append(token.pos_)
        tag_vals.append(token.tag_)
        lem = token.lemma_
        lem_vals.append(lem if lem.isalpha() else None)
        case = None
        if token.is_lower:
            case = "lower"
        elif token.is_title:
            case = "title"
        elif token.is_upper:
            case = "upper"
        elif token.is_space:
            case = "space"
        elif token.is_punct:
            case = "punct"
        elif token.is_quote:
            case = "quote"
        elif token.is_digit:
            case = "number"
        elif token.is_bracket:
            case = "bracket"
        elif token.is_currency:
            case = "currency"
        case_vals.append(case)

    cell.append(
        (
            tok_spans,
            {
                "pos": pos_vals,
                "tag": tag_vals,
                "lemma": lem_vals,
                "case": case_vals,
            },
            "TOKS",
        )
    )

    # Entities
    ent_spans = []
    ent_vals = []
    for ent in doc.ents:
        ent_spans.append((ent.start_char, ent.end_char))
        ent_vals.append(ent.label_)
    cell.append((ent_spans, {"entity": ent_vals}))

    # Sentences and noun chunks
    cell.append(
        ([(sent.start_char, sent.end_char) for sent in doc.sents], {}, "SENTS")
    )
    cell.append(
        (
            [(chunk.start_char, chunk.end_char) for chunk in doc.noun_chunks],
            {},
            "NOUNCH",
        )
    )

    return cell


def spacy_atterize_fn(cell, spacy_atterize_, nlp):
    # Adding spacytextblob, cannot do it in load_spacy because of
    # our multiprocessing code. Adding a pipe to SpaCy is idempotent.
    from spacytextblob.spacytextblob import (  # pylint: disable=unused-import
        SpacyTextBlob,
    )

    if not nlp.has_pipe("spacytextblob"):
        nlp.add_pipe("spacytextblob")
    return spacy_atterize_(nlp(cell))


def load_spacy():
    import spacy

    # nlp = spacy.load("en_core_web_sm", exclude=["parser"])
    nlp = spacy.load("en_core_web_sm")
    nlp.enable_pipe("senter")
    return nlp


def flair_atterize(sent):

    enriched_cell = []

    ent_spans = []
    ent_values = []
    ent_scores = []
    for ent in sent.get_spans("ner"):
        ent_spans.append((ent.start_position, ent.end_position))
        ent_values.append(ent.get_label("ner").value)
        ent_scores.append(str(int(round(ent.get_label("ner").score, 2) * 100)))
    enriched_cell.append(
        (ent_spans, {"entity": ent_values, "score": ent_scores}, "ENTS")
    )

    return enriched_cell


def flair_atterize_fn(cell, flair_atterize_, tagger_pred, sent_fn):
    sent = sent_fn(cell)
    tagger_pred(sent)
    return flair_atterize_(sent)


def load_flair():
    from flair.data import Sentence
    from flair.models import SequenceTagger

    # import logging
    # import warnings

    # logging.getLogger("flair").setLevel(logging.ERROR)
    # warnings.filterwarnings("ignore", module="huggingface_hub")

    tagger = SequenceTagger.load("ner")
    return (tagger.predict, Sentence)


def enrich_row(row):
    """
    Processes one row. Takes an input row and returns an output row. The
    global `ENRICHMENT_ARGS` would have previously been set.
    """
    atterize_fn = ENRICHMENT_ARGS[0]
    atterize_args = ENRICHMENT_ARGS[1:]

    cells = []

    for raw_cell in row:
        raw_cell = str(raw_cell)

        cell = atterize_fn(raw_cell, *atterize_args)
        adjust_span_offsets_from_char_to_byte(raw_cell, cell)

        cells.append(cell)

    return cells


def adjust_span_offsets_from_char_to_byte(raw_cell, cell):
    """
    Adjust all spans from character offsets to byte offsets, since Watchful's
    attributes API takes byte offsets and spacy provides us with character
    offsets.
    """
    byte_offsets = {}
    byte_offset = 0

    for char_offset, ch in enumerate(raw_cell):
        byte_offsets[char_offset] = byte_offset
        byte_offset += len(ch.encode("utf-8"))
    byte_offsets[len(raw_cell)] = byte_offset

    for context in cell:
        spans = context[0]
        for i, span in enumerate(spans):
            (start, end) = span
            spans[i] = (byte_offsets[start], byte_offsets[end])


def init_args(
    *args,
) -> None:
    """
    In this function, we initialize the user arguments as global variables.
    Generally, you should not need to edit this function.

    Initializes a per-process context with the user function that will be used
    in the imap. This is not necessarily thread-safe but is multiprocess-safe.
    """

    global ENRICHMENT_ARGS
    ENRICHMENT_ARGS = args


def enrich(
    in_file: str,
    out_file: str,
    enrich_row_fn: Callable,
    enrichment_args,
) -> None:
    """
    Usage:
    in_file: filepath of the csv formatted original dataset
    out_file: filepath where the enriched attributes in Watchful custom format
    for Watchful client ingestion are saved to
    enrich_row_fn: user custom function for enriching every row of the dataset
    enrichment_args: additional user variables to perform the data enrichment

    Generally, you should not need to edit this function.
    """

    sys_encoding = sys.getdefaultencoding()

    with open(in_file, "r", encoding=sys_encoding) as infile:
        in_reader = csv.reader(infile)
        n_cols = len(next(in_reader))
        n_rows = None
        for n_rows, _ in enumerate(in_reader, 1):
            pass

    with open(in_file, "r", encoding=sys_encoding) as infile, open(
        out_file, "w", encoding=sys_encoding
    ) as outfile:

        in_reader = csv.reader(infile)
        next(in_reader)

        global ATTR_WRITER
        ATTR_WRITER = writer(outfile, n_rows, n_cols)

        if IS_MULTIPROC:
            # Parallelize to the number of available cores (not the number of
            # available hyper threads).  `psutil` is the only standard Python
            # package that can provide this measure (with logical=False).
            # Testing revealed wall times to be quite close to using all logical
            # CPUs, with better overall system responsiveness and less thermal
            # throttling in this scenario.
            # Additionally, as Python's threading uses a GIL, it is unsuitable
            # for this task.  Use its multiprocessing intsead.  However,
            # multiprocessing uses pickle and is unable to send functions across
            # process boundaries, hence the global variable set by the
            # initializer.
            with Pool(
                initializer=init_args,
                initargs=enrichment_args,
                processes=psutil.cpu_count(logical=False),
            ) as pool:
                for enriched_row in pool.imap(
                    func=enrich_row_fn,
                    iterable=in_reader,
                    chunksize=MULTIPROC_CHUNKSIZE,
                ):
                    proc_enriched_row(enriched_row)
        else:
            init_args(*enrichment_args)
            for enriched_row in map(enrich_row_fn, in_reader):
                proc_enriched_row(enriched_row)

        del ATTR_WRITER


def proc_enriched_row(
    enriched_row: List[EnrichedCell],
) -> None:
    """
    This function is iterated over every enriched row. Optionally, you may add
    code if you wish to do something else with every row.
    """

    ########################################
    # Do not edit this code.
    for enriched_cell in enriched_row:
        proc_enriched_cell(enriched_cell)
    ########################################

    ########################################
    # EDIT THE CODE BELOW if you wish to do
    # something else with every row. Here
    # we simply print to stdout at the end
    # of every `enriched_row`.
    # print(f'{"*" * 20} end of row {"*" * 20}')
    ########################################


def proc_enriched_cell(
    enriched_cell: EnrichedCell,
) -> None:
    """
    This function is iterated over every enriched cell. Optionally, you may add
    code if you wish to do something else with every cell.
    """

    ########################################
    # Do not edit this code.
    ATTR_WRITER(enriched_cell)
    ########################################

    ########################################
    # EDIT THE CODE BELOW if you wish to do
    # something else with every cell. Here
    # we simply print every `enriched_cell`
    # to stdout.
    # print("Enriched cell: ")
    # pprint(enriched_cell)
    ########################################


def get_vars_for_enrich_row_with_attribute_data(attr_names, attr_file):

    f = open(  # pylint: disable=consider-using-with
        attr_file, "r", encoding=sys.getdefaultencoding()
    )
    attr_reader = csv.reader(f)
    attr_name_list_all = next(attr_reader)

    if not attr_names:
        attr_name_list = attr_name_list_all

        def get_attr_row(attr_row_all):
            return attr_row_all

    else:
        attr_name_list = attr_names.split(",")
        attr_locs = list(map(attr_name_list_all.index, attr_name_list))

        def get_attr_row(attr_row_all):
            return list(map(attr_row_all.__getitem__, attr_locs))

    return (get_attr_row, attr_name_list, attr_reader)


def enrich_row_with_attribute_data(row):
    """
    Extracts attributes from an attributes file. Attributes are associated to
    the entire text in each input dataset cell. The entire text in each input
    dataset cell is identified by its character start index and character end
    index.
    """

    get_attr_row_fn = ENRICHMENT_ARGS[0]
    attr_names = ENRICHMENT_ARGS[1]
    attr_row_reader = ENRICHMENT_ARGS[2]

    attr_row = get_attr_row_fn(next(attr_row_reader))
    cells_attrs = []

    for in_cell in row:
        # Just one span for example-level attributes.
        span = [(0, len(str(in_cell)))]

        # `cell_attrs` is to be appended over the len of attr_names.
        cell_attrs = []

        for attr_name, attr_val in zip(attr_names, attr_row):
            # This is the attribute representation per input dataset cell as
            # required by Watchful client interface. Watchful client has a
            # restriction of uppercase characters and maximum 6 characters on
            # attr_representn[2].
            # attr_representn = (
            #     span, {attr_name: [attr_val]},
            #     f"EX{attr_name.replace("_", "").upper()}"
            # )
            # attr_representn = (
            #     span, {attr_name: [attr_val]},
            #     "SOME_GENERAL_INFO_ABOUT_ALL_ATTRIBUTES"
            # )
            attr_representn = (span, {attr_name: [attr_val]})
            cell_attrs.append(attr_representn)

        # Another possible representation.
        # attr_dict = {}
        # for attr_name, attr_val in zip(attr_names, attr_row):
        #     attr_dict[attr_name] = [attr_val]
        # attr_representn = (span, attr_dict, "EXATTR")
        # cell_attrs.append(attr_representn)

        cell_attrs.append((span, {}, "EX"))

        cells_attrs.append(cell_attrs)

    return cells_attrs


def validate_attr_names(attr_names, attr_file):
    """
    Check that all attribute names are present in the attribute file. Returns
    `False` as soon as an attribute name is absent, or `True` when all attribute
    names match.
    """

    with open(attr_file, "r", encoding=sys.getdefaultencoding()) as f:
        attr_row_reader = csv.reader(f)
        attr_names_all = next(attr_row_reader)

    for attr_name in attr_names.split(","):
        if attr_name not in attr_names_all:
            return False

    return True


def atterize_values_in_cell(cell, name, values):
    """
    Helper function to `create_attribute_for_values()` for finding the spans for
    each value.
    """

    from heapq import merge

    cell = str(cell)
    matches = [
        [(m.start(), m.end()) for m in re.finditer(pattern, cell)]
        for pattern in values
    ]
    spans = list(merge(*matches))
    return [(spans, {}, name)]


def create_attribute_for_values(attribute_name, values):
    """
    Takes an attribute name and a list of known values to create attributes for.
    The list of values will be looked up in each cell of the currently loaded
    dataset. An attributes file will be prepared to be loaded in watchful.
    Returns the filename of the attribute file created, which can be used by the
    attributes action and function:
    api.load_attributes(dataset_id, attribute_filename)
    """

    in_file, out_file, out_filename = get_context(attribute_name)
    enrich(
        in_file,
        out_file,
        enrich_row,
        (atterize_values_in_cell, attribute_name, values),
    )
    return out_filename


def get_context(attribute_filename):
    """
    Given a `attribute_filename`, find the current dataset file loaded in
    Watchful and produce the context needed to enrich that dataset. Also returns
    the filename of the file used by the attributes action and function:
    `api.load_attributes(dataset_id, attribute_filename)`
    """

    summary = client.get()
    attrs_dir = os.path.join(summary["watchful_home"], "datasets", "attrs")
    os.makedirs(attrs_dir, exist_ok=True)
    _, _, in_file = get_dataset_id_dir_filepath(summary)
    in_filename = os.path.basename(in_file)
    out_file = os.path.join(
        attrs_dir, f"{in_filename}_{attribute_filename}.attrs"
    )
    out_filename = os.path.basename(out_file)
    return in_file, out_file, out_filename


def get_dataset_id_dir_filepath(summary, in_file="", is_local=True):
    """
    This function returns the id, directory and filepath of the currently opened
    dataset.
    """

    summary = client._assert_success(summary)
    dataset_id = client.get_dataset_id(summary)
    datasets_dir = client.get_datasets_dir(summary, is_local)

    if in_file != "":
        # Check that `in_file` exists.
        if not os.path.isfile(in_file):
            raise Exception(f"File {in_file} does not exist.")
        dataset_filepath = in_file
    else:
        dataset_filepath = client.get_dataset_filepath(summary, is_local)

    return dataset_id, datasets_dir, dataset_filepath



"""
This script provides the functions required for interacting directly with
Watchful client.
"""
################################################################################


# TODO:
# - We need a better way of signalling error besides making the caller check for
#   it in the summary. Maybe throw from api() if error_msg is set?


import csv
import http.client
import io
import json
import os
import socket
import subprocess
import sys
import time
import urllib


HOST = "localhost"
PORT = "9002"


def _refresh():
    del sys.modules["watchful"]
    import watchful

    print(f"watchful version: {watchful.__version__}")


def _get_conn():
    return http.client.HTTPConnection(HOST + ":" + PORT)


def await_port_opening(port, timeout_sec=10):
    """
    Waits for the port to be open.
    Returns true if the port was open within the timeout.
    Used for awaiting Watchful process startup.
    """
    end = time.time_ns() + (timeout_sec * 1_000_000_000)
    while time.time_ns() < end:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout_sec)
        res = sock.connect_ex((HOST, port))
        if res == 0:
            return
        time.sleep(0.001)
    raise Exception("Timed out waiting for Watchful to start")


def spawn_cmd(cmd, env=None):
    """Spawns a command and returns the PID of the spawned process"""
    with subprocess.Popen(
        cmd + " &\n echo $!", shell=True, stdout=subprocess.PIPE, env=env
    ) as proc:
        out, _ = proc.communicate()
    pid = int(out.decode("utf-8")[:-1])
    return pid


def await_summary(pred, halt_fn=lambda x: False, unchanged_timeout=60):
    """
    Returns the summary once predicate(summary) returns true.
    Stops waiting once the halt function returns true, and returns None.
    Raises an exception if the summary is unchanged for `unchanged_timeout`
    seconds.
    """
    prev_summary = None
    end = float("inf")
    while time.time_ns() < end:
        summary = get()
        if halt_fn(summary):
            return None
        if pred(summary):
            return summary
        if prev_summary == summary:
            end = time.time_ns() + (unchanged_timeout * 1_000_000_000)
        prev_summary = summary
        time.sleep(0.03)  # 30ms sleep for ~30 requests per second

    raise Exception(
        "Timed out awaiting summary. Summary went stale for "
        + str(unchanged_timeout)
        + " seconds"
    )


def _assert_success(summary):
    if "error_msg" in summary and summary["error_msg"]:
        verb_str = ""
        if "error_verb" in summary and summary["error_verb"]:
            verb_str = f' ({summary["error_verb"]})'
        raise Exception(f'Summary error{verb_str}: {summary["error_msg"]}')
    return summary


def _read_response_summary(resp):
    assert 200 == int(resp.status)
    ret = json.loads(resp.read())
    # One idea:
    # if ret["error_msg"]:
    #     raise Exception(ret["error_msg"])
    if "error_msg" in ret and ret["error_msg"]:
        print(ret["error_msg"])
    return ret


def api(verb, **args):
    """
    Convenience method for API calls; made up of a verb and some arguments.
    """
    action = args  # already a dictionary
    action["verb"] = verb
    return api_send_action(action)


def api_send_action(action):
    """
    Convenience method used by `se_auto_evaluator.py`. Consequently, api()
    should be refactored to reuse this function abstraction. Why? Where `action`
    is directly available, we should not strip it for api() and reform it in
    api().
    """
    conn = _get_conn()
    conn.request(
        "POST", "/api", json.dumps(action), {"Content-Type": "application/json"}
    )
    return _read_response_summary(conn.getresponse())


def ephemeral(port="9002"):
    """Start the backend for an interactive session without persistence."""
    spawn_cmd(
        f"watchful -p {port} --no-persistence >watchful_ephemeral_output.txt 2>&1"
    )
    await_port_opening(int(port))
    external(port=port)


def external(host="localhost", port="9001"):
    global HOST, PORT
    HOST = host
    PORT = port


def list_projects():
    c = _get_conn()
    c.request("GET", "/projects")
    r = c.getresponse()
    return json.loads(r.read())


def open_project(id_):
    c = _get_conn()
    c.request(
        "POST",
        "/projects",
        json.dumps(id_),
        {"Content-Type": "application/json"},
    )
    r = c.getresponse()
    ret = r.read()
    await_plabels()
    return ret


def create_project():
    return open_project("new")


def get_project_id(summary):
    """
    Get the project id from the summary. For correctness, we use the summary
    that has been success asserted.
    """
    if "project_id" in summary:
        return summary["project_id"]
    raise Exception("No project is currently active.")


def get_dataset_id(summary):
    """
    Get the dataset id from the summary. For correctness, we use the summary
    that has been success asserted.
    """
    if "datasets" in summary:
        # `dataset_ids` should either be empty or contain one dataset id.
        dataset_ids = summary["datasets"]
        if len(dataset_ids) == 0:
            raise Exception("No dataset is currently opened.")
        dataset_id = dataset_ids[0]
        return dataset_id
    raise Exception("`datasets` is currently not available.")


def get_watchful_home(summary, is_local=True):
    """
    Get Watchful home from the summary. For correctness, we use the summary that
    has been success asserted. If Watchful home is not available and the
    Watchful application is local, we derive Watchful home from the user home.
    """
    if "watchful_home" in summary:
        return summary["watchful_home"]
    if is_local:
        user_home = os.path.expanduser("~")
        watchful_home = os.path.join(user_home, "watchful")
        return watchful_home
    raise Exception("`watchful_home` is currently not available.")


def get_datasets_dir(summary, is_local=True):
    """
    Infer the datasets directory from the summary. For correctness, we use the
    summary that has been success asserted.
    """
    watchful_home = get_watchful_home(summary, is_local)
    datasets_dir = os.path.join(watchful_home, "datasets")
    return datasets_dir


def get_dataset_filepath(summary, is_local=True):
    """
    Infer the dataset filepath from the summary. For correctness, we use the
    summary that has been success asserted. As this function uses file
    operations, it does not work when the Watchful application is remote, and in
    this case returns "".
    """
    if not is_local:
        return ""

    datasets_dir = get_datasets_dir(summary, is_local)
    dataset_id = get_dataset_id(summary)
    dataset_ref_path = os.path.join(datasets_dir, "refs", dataset_id)

    # Check that `dataset_ref_path` exists.
    if not os.path.isfile(dataset_ref_path):
        raise Exception(f"File {dataset_ref_path} does not exist.")
    with open(dataset_ref_path, encoding=sys.getdefaultencoding()) as f:
        dataset_ref = f.readline()
    dataset_filepath = os.path.join(datasets_dir, "raw", dataset_ref)

    # Check that `dataset_filepath` exists.
    if not os.path.isfile(dataset_filepath):
        raise Exception(f"File {dataset_filepath} does not exist.")

    return dataset_filepath


# def poll():


def records(csv_):
    return api("records", data=csv_, content_type="text/csv")


def class_(class__):
    return api("class", name=class__)


def create_class(class__, class_type="ftc"):
    return api("class", name=class__, class_type=class_type)


def query_async(q, page=0):
    return api("query", query=q, page=page)


def query(q, page=0):
    query_async(q, page)
    return await_summary(
        lambda s: s["query_completed"], lambda s: s["query"] != q
    )


def query_all(q, max_pages=0):
    """
    Evaluates the query returning the results as opposed to the summary. Will by
    default return all results for the query (all pages). This can be limited by
    setting max_pages to the positive number of pages you want.
    Each query result is a vector with a string for each field that is returned.
    Note: TOKS, SENTS, CELLS queries only return one field and each result will
    be wrapped in a vector of one string.
    """
    page = 0
    while True:
        summary = query(q, page)
        for fields in [cand["fields"] for cand in summary["candidates"]]:
            yield fields
        page += 1
        if summary["query_end"] or max_pages and page == max_pages:
            break


def base_rate(class__, rate):
    return api("base_rate", label=class__, rate=rate)


def await_plabels():
    return await_summary(lambda s: s["status"] == "current")


def hinter_async(class__, query_, weight):
    return api("hinter", label=class__, weight=weight, query=query_)


def hinter(class__, query_, weight):
    _assert_success(hinter_async(class__, query_, weight))
    return await_plabels()


def delete(id_):
    return api("delete", id=id_)


def delete_class(class__):
    return api("delete", class_name=class__)


def get():
    return api("nop")


def title(title_):
    return api("title", title=title_)


def external_hinter(class__, name, weight):
    _assert_success(
        api(
            "hinter",
            query="[external]",
            name=name,
            label=class__,
            weight=weight,
        )
    )
    return await_plabels()


def upload_attributes(dataset_id, attributes_filepath):
    """
    This function is used in the case of Watchful application being on a
    machine different from where data enrichment is run.
    Note: Eventually upload_attributes() should upload the attributes to the
    Watchful application via the api and saves it to a filepath according to
    the application logic. With this, upload_attributes() can be used for both
    local and remote Watchful application.
    """
    conn = _get_conn()
    with open(attributes_filepath, "r", encoding="utf-8") as attr_file:
        conn.request(
            "PUT",
            f"/datasets/{dataset_id}/attributes",
            body=attr_file,
            headers={"Content-Type": "text/plain"},
        )
    resp = conn.getresponse()
    assert resp.status == 200, f"not OK HTTP status. Was: {resp.status}"
    return _assert_success(_read_response_summary(resp))


def load_attributes(dataset_id, attributes_filename):
    """
    This function is used in the case of Watchful application being on the same
    machine as the data enrichment. Loads attributes into the local Watchful
    using dataset_id:string and attributes_filename:string.
    """
    return api("attributes", id=dataset_id, file=attributes_filename)


def _dump(offset):
    return api("dump", offset=offset)


def dump():
    """
    A generator that returns all the candidates in "hint API order".
    """
    n_cands = get()["n_candidates"]
    offset = 0
    chunk = []
    while offset < n_cands:
        summary = _dump(offset)
        # TODO: handle errors
        chunk = summary["candidates"]
        offset += len(chunk)
        i = 0
        while i < len(chunk):
            yield chunk[i]
            i += 1


def dump_dicts():
    field_names = get()["field_names"]
    for c in dump():
        yield dict(zip(field_names, c))


def hint(name, offset, values):
    """
    TODO: Come up with a better streaming Python API here.
    """
    values = list(map(lambda x: x and 1 or 0, values))
    return _assert_success(api("hint", name=name, offset=offset, values=values))


def apply_hints(name):
    _assert_success(api("apply_hints", name=name))
    return await_plabels()


def hint_all(name, values):
    hint(name, 0, values)
    return apply_hints(name)


def export_stream(content_type="text/csv", mode="ftc"):
    """
    Begins the export using the export_stream call.  The result is not
    JSON, but is data to be processed directly.

    For FTC mode, content_type must be text/csv and mode must be ftc.

    For NER mode, content_type must be application/jsonlines and mode
    must be ner.

    On success, returns the HTTPResponse object from which you can
    read the data.
    """
    conn = _get_conn()
    conn.request(
        "GET",
        "/export_stream?content-type="
        + urllib.parse.quote_plus(content_type)
        + "&mode="
        + urllib.parse.quote_plus(mode),
    )
    resp = conn.getresponse()
    assert 200 == int(resp.status)
    return resp


def export_dataset_to(out_file, fields=None):
    """
    Exports the original dataset via a buffered stream to the specified
    output file.
    Takes header as an optional argument, for when the callee expects
    a certain header.

    :raise exception when the exported file does not match the expected header.
    """
    if not fields:
        fields = get()["field_names"]
    n_cols = len(fields)
    dataset_export_stream = export_stream(content_type="text/csv", mode="ftc")
    with open(out_file, "w", encoding="utf-8") as f:
        writer = csv.writer(f)
        reader = csv.reader(
            io.TextIOWrapper(
                dataset_export_stream, encoding="utf-8", newline=""
            )
        )
        header = next(reader)
        if header[:n_cols] != fields:
            raise Exception(
                f"Dataset did not match expected field names. header={header} != fields={fields}"
            )
        writer.writerow(fields)
        for row in reader:
            writer.writerow(row[:n_cols])


def export_async():
    return api("export", query="", content_type="text/csv")


def export():
    n_exports = len(get()["exports"])
    export_async()
    return await_summary(lambda s: len(s["exports"]) == n_exports + 1)


def export_preview(mode="ftc"):
    """
    Returns a preview of the export on summary key 'export_preview'
    The mode parameter can be "ftc" or "ner".
    """
    return api("export_preview", mode=mode)


def create_dataset(csv_bytes, columns, filename="none", has_header=True):
    """
    TODO: Add error handling
    """
    from uuid import uuid4

    id_ = str(uuid4())
    conn = _get_conn()
    conn.request(
        "POST",
        "/api/_stream/" + id_ + "/0/true",
        csv_bytes,
        {"Content-Type": "text/csv"},
    )
    _ = _read_response_summary(conn.getresponse())
    params = json.dumps({"filename": filename, "has_header": has_header})
    conn.request(
        "POST",
        "/api/_stream/" + id_,
        params,
        {"Content-Type": "application/json"},
    )
    dataset_id = _read_response_summary(conn.getresponse())["id"]
    api("dataset_add", id=dataset_id, columns=columns)
    return dataset_id


def label_single(row):
    c = _get_conn()
    sio = io.StringIO()
    w = csv.writer(sio)
    w.writerow(row)
    csv_row = sio.getvalue().encode("utf-8")
    sio.close()
    c.request("POST", "/label", csv_row, {"Content-Type": "text/csv"})
    r = c.getresponse()
    resp_body = r.read().decode("utf-8")
    csv_str = io.StringIO(resp_body)
    rdr = csv.reader(csv_str)
    rdr_list = list(rdr)
    if len(rdr_list) == 2:
        fields = get()["field_names"]
        assert (
            fields == rdr_list[0][0 : len(fields)]
        ), "server prepended the header to the labeled row"
    else:
        assert len(rdr_list) == 1, "server returned a single row"
    return rdr_list[-1]


def config_set(key, value):
    conn = _get_conn()
    params = json.dumps({"verb": "set", "key": key, "value": value})
    conn.request(
        "POST", "/config", params, {"Content-Type": "application/json"}
    )
    return _read_response_summary(conn.getresponse())


def config():
    conn = _get_conn()
    conn.request("GET", "/config", None, {"Content-Type": "application/json"})
    return _read_response_summary(conn.getresponse())


def print_candidates(summary=None):
    if summary is None:
        summary = get()
    print(
        "\n".join(
            [",".join(summary["field_names"])]
            + list(map(lambda c: ",".join(c["fields"]), summary["candidates"]))
        )
    )


def candidate_dicts(summary=None):
    if summary is None:
        summary = get()
    return list(
        map(
            lambda c: dict(zip(summary["field_names"], c["fields"])),
            summary["candidates"],
        )
    )


def exit_backend():
    """
    Note the API call will usually fail because the backend exits before
    returning an HTTP response so we suppress the error.
    """
    try:
        api("exit")
    # We allowing passing this error as it is known.
    except http.client.RemoteDisconnected:
        pass


### Hub API


def hub_api(verb, token, **args):
    """
    Convenience method for collaboration API calls.  Makes a hub
    API request of the backend.
    """
    headers = {"Content-Type": "application/json"}
    headers.update({"Authorization": "Bearer " + token})
    action = args  # already a dictionary
    action["verb"] = verb
    conn = _get_conn()
    conn.request("POST", "/remote", json.dumps(action), headers)
    return _assert_success(_read_response_summary(conn.getresponse()))


def login(email, password):
    import base64

    headers = {"Content-Type": "application/json"}
    credentials = base64.b64encode(str.encode(f"{email}:{password}")).decode(
        "utf-8"
    )
    headers.update({"Authorization": f"Basic {credentials}"})
    conn = _get_conn()
    conn.request("POST", "/remote", json.dumps({"verb": "login"}), headers)
    return _assert_success(_read_response_summary(conn.getresponse()))


def publish(token=None):
    return hub_api("publish", token)


def fetch(token=None):
    return hub_api("fetch", token)


def pull(token=None):
    return hub_api("pull", token)


def push(token=None):
    return hub_api("push", token)


def peek(token=None):
    return hub_api("peek", token)


def whoami(token=None):
    return hub_api("whoami", token)



"""
Initializes "watchful" as a module.
"""
################################################################################


import os
import sys

# We need to be careful of unintended overriding, and also a deeper
# consideration into what should be made directly available in the "watchful"
# namespace.
from .client import *
from .attributes import *
from .enrich import main
from .enricher import *


THIS_FILE_DIR = os.path.dirname(os.path.abspath(__file__))
with open(
    os.path.join(THIS_FILE_DIR, "VERSION"),
    "r",
    encoding=sys.getdefaultencoding(),
) as f:
    __version__ = f.readline()



"""
This script provides the template ``Enricher`` class interface for adding your
custom data enrichment functions and models. Some example enricher classes are
provided below for your reference.
"""
################################################################################


from abc import ABCMeta, abstractmethod
from typing import Generic, Iterable, List, TypeVar
from watchful import attributes


class Enricher(metaclass=ABCMeta):
    """
    This is the interface that customized enricher classes should inherit, and
    then implement the abstract methods.
    """

    @abstractmethod
    def __init__(
        self,
    ) -> None:
        """
        In this function, we create variables that we will later use in
        `enrich_row` to enrich our data row-wise.
        """
        pass

    @abstractmethod
    def enrich_row(
        self,
        row: Iterable[str],
    ) -> List[attributes.EnrichedCell]:
        """
        In this function, we use our variables from `self.enrichment_args` to
        enrich every row of your data. The return value is our enriched row.
        """
        pass

    @classmethod
    def is_enricher(
        cls,
        possibly_an_enricher: Generic[TypeVar("T")],
    ) -> bool:
        """
        Convenience method for checking if `possibly_an_enricher` is indeed of
        type `Enricher`.
        """
        return issubclass(possibly_an_enricher, cls)
